package gregapi.block.multitileentity;

import static gregapi.data.CS.W;
import gregapi.block.IBlockMaterial;
import gregapi.block.IBlockRetrievable;
import gregapi.block.IBlockToolable;
import gregapi.code.ArrayListNoNulls;
import gregapi.oredict.OreDictMaterialStack;
import gregapi.render.IRenderedBlock;
import gregapi.render.IRenderedBlockObject;
import gregapi.render.ITexture;
import gregapi.render.RendererBlockTextured;
import gregapi.tileentity.ITileEntity;
import gregapi.util.UT;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.minecraft.block.Block;
import net.minecraft.block.ITileEntityProvider;
import net.minecraft.block.material.Material;
import net.minecraft.client.renderer.texture.IIconRegister;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.enchantment.EnchantmentHelper;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.Item;
import net.minecraft.item.ItemBlock;
import net.minecraft.item.ItemStack;
import net.minecraft.stats.StatList;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.IIcon;
import net.minecraft.util.MovingObjectPosition;
import net.minecraft.util.StatCollector;
import net.minecraft.world.Explosion;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.common.util.ForgeDirection;
import net.minecraftforge.event.ForgeEventFactory;
import cpw.mods.fml.common.registry.GameRegistry;

/**
 * @author Gregorius Techneticies
 */
public class MultiTileEntityBlock extends Block implements IRenderedBlock, ITileEntityProvider, IBlockToolable, IBlockRetrievable, IBlockMaterial {
	/** There are quite some odd timings in the Block breaking Code which require the TileEntity to be buffered. */
	public static ThreadLocal<TileEntity> sTemporaryTileEntity = new ThreadLocal<TileEntity>();
	
	public final int mHarvestLevelOffset, mHarvestLevelMinimum, mHarvestLevelMaximum;
	public final String mNameInternal, mTool;
	public final boolean mOpaque, mNormalCube;
	
	public static Map<String, MultiTileEntityBlock> MULTITILEENTITYBLOCKMAP = new HashMap<String, MultiTileEntityBlock>();
	
	public static String getName(String aNameOfVanillaMaterialField, Material aVanillaMaterial, SoundType aSoundType, String aTool, int aHarvestLevelOffset, int aHarvestLevelMinimum, int aHarvestLevelMaximum, boolean aOpaque, boolean aNormalCube) {
		return "gt.multitileentityblock." + aNameOfVanillaMaterialField + "." + aSoundType.soundName + "." + aTool + "." + aHarvestLevelOffset + "." + aHarvestLevelMinimum + "." + aHarvestLevelMaximum + "." + aOpaque + "." + aNormalCube;
	}
	
	public static MultiTileEntityBlock getOrCreate(String aNameOfVanillaMaterialField, Material aVanillaMaterial, SoundType aSoundType, String aTool, int aHarvestLevelOffset, int aHarvestLevelMinimum, int aHarvestLevelMaximum, boolean aOpaque, boolean aNormalCube) {
		MultiTileEntityBlock rBlock = MULTITILEENTITYBLOCKMAP.get(getName(aNameOfVanillaMaterialField, aVanillaMaterial, aSoundType, aTool = aTool.toLowerCase(), aHarvestLevelOffset, aHarvestLevelMinimum, aHarvestLevelMaximum, aOpaque, aNormalCube));
		return rBlock == null ? new MultiTileEntityBlock(aNameOfVanillaMaterialField, aVanillaMaterial, aSoundType, aTool, aHarvestLevelOffset, aHarvestLevelMinimum, aHarvestLevelMaximum, aOpaque, aNormalCube) : rBlock;
	}
	
	/**
	 * @param aNameOfVanillaMaterialField the Name of the vanilla Material Field. In case this is not a vanilla Material, insert the Name you want to give your own Material instead.
	 * @param aVanillaMaterial the Material used to determine the Block.
	 * @param aSoundType the Sound Type of the Block.
	 * @param aTool the Tool used to harvest this Block.
	 * @param aHarvestLevelOffset
	 * @param aHarvestLevelMinimum
	 * @param aHarvestLevelMaximum
	 * @param aOpaque if this Block is Opaque.
	 * @param aNormalCube if this Block is a normal Cube (for Redstone Stuff).
	 */
	private MultiTileEntityBlock(String aNameOfVanillaMaterialField, Material aVanillaMaterial, SoundType aSoundType, String aTool, int aHarvestLevelOffset, int aHarvestLevelMinimum, int aHarvestLevelMaximum, boolean aOpaque, boolean aNormalCube) {
		super(aVanillaMaterial);
		mNameInternal = getName(aNameOfVanillaMaterialField, aVanillaMaterial, aSoundType, aTool, aHarvestLevelOffset, aHarvestLevelMinimum, aHarvestLevelMaximum, aOpaque, aNormalCube);
		GameRegistry.registerBlock(this, ItemBlock.class, mNameInternal);
		MULTITILEENTITYBLOCKMAP.put(mNameInternal, this);
		
		setStepSound(aSoundType);
		mOpaque = aOpaque;
		mNormalCube = aNormalCube;
		
		mTool = aTool.toLowerCase();
		mHarvestLevelOffset = aHarvestLevelOffset;
		mHarvestLevelMinimum = Math.max(0, aHarvestLevelMinimum);
		mHarvestLevelMaximum = Math.max(aHarvestLevelMinimum, aHarvestLevelMaximum);
		
        opaque = this.isOpaqueCube();
        lightOpacity = this.isOpaqueCube() ? 255 : 0;
        
        try {codechicken.nei.api.API.hideItem(new ItemStack(this, 1, W));} catch(Throwable e) {/**/}
	}
	
	private static boolean LOCK = false;
	
    // You want to override one of those Functions with your TileEntity? Just implement the Interfaces on your TileEntity and you are done.
    
    public static interface IMultiTileEntityOnBlockExploded {public void onExploded(Explosion aExplosion);}
    @Override public final void onBlockExploded(World aWorld, int aX, int aY, int aZ, Explosion aExplosion) {if (aWorld.isRemote) return; TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); if (aTileEntity instanceof IMultiTileEntityOnBlockExploded) ((IMultiTileEntityOnBlockExploded)aTileEntity).onExploded(aExplosion); else aWorld.setBlockToAir(aX, aY, aZ);}
	
    public static interface IMultiTileEntityGetPickBlock {public ItemStack getPickBlock(MovingObjectPosition aTarget);}
    @Override public final ItemStack getPickBlock(MovingObjectPosition aTarget, World aWorld, int aX, int aY, int aZ) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetPickBlock?((IMultiTileEntityGetPickBlock)aTileEntity).getPickBlock(aTarget):null;}
	
    public static interface IMultiTileEntityBreakBlock {public void breakBlock();}
    @Override public final void breakBlock(World aWorld, int aX, int aY, int aZ, Block aBlock, int par6) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); sTemporaryTileEntity.set(aTileEntity); if (aTileEntity instanceof IMultiTileEntityBreakBlock) ((IMultiTileEntityBreakBlock)aTileEntity).breakBlock(); else aWorld.removeTileEntity(aX, aY, aZ);}
	
    public static interface IMultiTileEntityGetStackFromBlock {public ItemStack getStackFromBlock(byte aSide);}
    @Override public final ItemStack getItemStackFromBlock(IBlockAccess aWorld, int aX, int aY, int aZ, byte aSide) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetStackFromBlock?((IMultiTileEntityGetStackFromBlock)aTileEntity).getStackFromBlock(aSide):null;}
    
    public static interface IMultiTileEntityGetFlammability {public int getFlammability(byte aSide);}
    @Override public final int getFlammability(IBlockAccess aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetFlammability ? ((IMultiTileEntityGetFlammability)aTileEntity).getFlammability((byte)aSide.ordinal()) : 0;}
	
    public static interface IMultiTileEntityGetFireSpreadSpeed {public int getFireSpreadSpeed(byte aSide);}
    @Override public final int getFireSpreadSpeed(IBlockAccess aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetFireSpreadSpeed ? ((IMultiTileEntityGetFireSpreadSpeed)aTileEntity).getFireSpreadSpeed((byte)aSide.ordinal()) : 0;}
	
    public static interface IMultiTileEntityIsFireSource {public boolean isFireSource(byte aSide);}
	@Override public final boolean isFireSource(World aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityIsFireSource ? ((IMultiTileEntityIsFireSource)aTileEntity).isFireSource((byte)aSide.ordinal()) : false;}
	
	public static interface IMultiTileEntityCanEntityDestroy {public boolean canEntityDestroy(Entity aEntity);}
	@Override public final boolean canEntityDestroy(IBlockAccess aWorld, int aX, int aY, int aZ, Entity aEntity) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityCanEntityDestroy ? ((IMultiTileEntityCanEntityDestroy)aTileEntity).canEntityDestroy(aEntity) : true;}
	
	public static interface IMultiTileEntityOnToolClick {public long onToolClick(String aTool, long aRemainingDurability, Entity aPlayer, IInventory aPlayerInventory, boolean aSneaking, ItemStack aStack, byte aSide, float aHitX, float aHitY, float aHitZ);}
	@Override public final long onToolClick(String aTool, long aRemainingDurability, Entity aPlayer, IInventory aPlayerInventory, boolean aSneaking, ItemStack aStack, World aWorld, byte aSide, int aX, int aY, int aZ, float aHitX, float aHitY, float aHitZ) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityOnToolClick ? ((IMultiTileEntityOnToolClick)aTileEntity).onToolClick(aTool, aRemainingDurability, aPlayer, aPlayerInventory, aSneaking, aStack, aSide, aHitX, aHitY, aHitZ) : 0;}
	
	public static interface IMultiTileEntityGetMaterialAtSide {public OreDictMaterialStack getMaterialAtSide(byte aSide);}
	@Override public final OreDictMaterialStack getMaterialAtSide(IBlockAccess aWorld, int aX, int aY, int aZ, byte aSide) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetMaterialAtSide?((IMultiTileEntityGetMaterialAtSide)aTileEntity).getMaterialAtSide(aSide):null;}
	
	public static interface IMultiTileEntityRemoveMaterialFromSide {public boolean removeMaterialFromSide(byte aSide, OreDictMaterialStack aMaterial);}
	@Override public final boolean removeMaterialFromSide(World aWorld, int aX, int aY, int aZ, byte aSide, OreDictMaterialStack aMaterial) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityRemoveMaterialFromSide ? ((IMultiTileEntityRemoveMaterialFromSide)aTileEntity).removeMaterialFromSide(aSide, aMaterial) : false;}
	
	public static interface IMultiTileEntityGetDrops {public ArrayListNoNulls<ItemStack> getDrops(int aFortune, boolean aSilkTouch);}
    @Override public final void dropBlockAsItemWithChance(World aWorld, int aX, int aY, int aZ, int aMeta, float aChance, int aFortune) {if (!aWorld.isRemote) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); if (aTileEntity == null) aTileEntity = sTemporaryTileEntity.get(); if (aTileEntity instanceof IMultiTileEntityGetDrops) {ArrayListNoNulls<ItemStack> tList = ((IMultiTileEntityGetDrops)aTileEntity).getDrops(aFortune, false); aChance = ForgeEventFactory.fireBlockHarvesting(tList, aWorld, this, aX, aY, aZ, aMeta, aFortune, aChance, false, harvesters.get()); for (ItemStack tStack : tList) if (aWorld.rand.nextFloat() <= aChance) dropBlockAsItem(aWorld, aX, aY, aZ, tStack);}}}
    @Override public final void harvestBlock(World aWorld, EntityPlayer aPlayer, int aX, int aY, int aZ, int aMeta) {aPlayer.addStat(StatList.mineBlockStatArray[getIdFromBlock(this)], 1); aPlayer.addExhaustion(0.025F); if (!aWorld.isRemote) {boolean aSilkTouch = EnchantmentHelper.getSilkTouchModifier(aPlayer); int aFortune = EnchantmentHelper.getFortuneModifier(aPlayer); float aChance = 1.0F; TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); if (aTileEntity == null) aTileEntity = sTemporaryTileEntity.get(); if (aTileEntity instanceof IMultiTileEntityGetDrops) {ArrayListNoNulls<ItemStack> tList = ((IMultiTileEntityGetDrops)aTileEntity).getDrops(aFortune, aSilkTouch); aChance = ForgeEventFactory.fireBlockHarvesting(tList, aWorld, this, aX, aY, aZ, aMeta, aFortune, aChance, aSilkTouch, harvesters.get()); for (ItemStack tStack : tList) if (aWorld.rand.nextFloat() <= aChance) dropBlockAsItem(aWorld, aX, aY, aZ, tStack);}}}
    @Override public final ArrayList<ItemStack> getDrops(World aWorld, int aX, int aY, int aZ, int aUnusableMetaData, int aFortune) {ArrayListNoNulls<ItemStack> tList = new ArrayListNoNulls<ItemStack>(); if (!aWorld.isRemote) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); if (aTileEntity == null) aTileEntity = sTemporaryTileEntity.get(); if (aTileEntity instanceof IMultiTileEntityGetDrops) tList = ((IMultiTileEntityGetDrops)aTileEntity).getDrops(aFortune, false);} return tList;}
    
    public static interface IMultiTileEntityGetBlockHardness {public float getBlockHardness(EntityPlayer aPlayer);}
    @Override public final float getPlayerRelativeBlockHardness(EntityPlayer aPlayer, World aWorld, int aX, int aY, int aZ) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetBlockHardness?Math.max(0.0F, ((IMultiTileEntityGetBlockHardness)aTileEntity).getBlockHardness(aPlayer)):1.0F;}
    @Override public final float getBlockHardness(World aWorld, int aX, int aY, int aZ) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetBlockHardness?Math.max(0.0F, ((IMultiTileEntityGetBlockHardness)aTileEntity).getBlockHardness(null)):1.0F;}
	
    public static interface IMultiTileEntityGetExplosionResistance {public float getExplosionResistance(Entity aExploder, double aExplosionX, double aExplosionY, double aExplosionZ);}
    @Override public final float getExplosionResistance(Entity aExploder, World aWorld, int aX, int aY, int aZ, double aExplosionX, double aExplosionY, double aExplosionZ) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetExplosionResistance?Math.max(0.0F, ((IMultiTileEntityGetExplosionResistance)aTileEntity).getExplosionResistance(aExploder, aExplosionX, aExplosionY, aExplosionZ)):1.0F;}
	
    public static interface IMultiTileEntityIsSideSolid {public boolean isSideSolid(byte aSide);}
    @Override public final boolean isSideSolid(IBlockAccess aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityIsSideSolid?((IMultiTileEntityIsSideSolid)aTileEntity).isSideSolid((byte)aSide.ordinal()):mOpaque;}
    
    public static interface IMultiTileEntityIsBeaconBase {public boolean isBeaconBase(int aBeaconX, int aBeaconY, int aBeaconZ);}
    @Override public final boolean isBeaconBase(IBlockAccess aWorld, int aX, int aY, int aZ, int aBeaconX, int aBeaconY, int aBeaconZ) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityIsBeaconBase?((IMultiTileEntityIsBeaconBase)aTileEntity).isBeaconBase(aBeaconX, aBeaconY, aBeaconZ):false;}
    
    public static interface IMultiTileEntityGetLightOpacity {public int getLightOpacity();}
    @Override public final int getLightOpacity(IBlockAccess aWorld, int aX, int aY, int aZ) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity instanceof IMultiTileEntityGetLightOpacity?((IMultiTileEntityGetLightOpacity)aTileEntity).getLightOpacity():mOpaque?255:0;}
    
    
    
	@Override public final boolean isOpaqueCube() {return mOpaque;}
	@Override public final boolean func_149730_j() {return mOpaque;}
	@Override public final boolean renderAsNormalBlock() {return mOpaque || mNormalCube;}
	@Override public final boolean isNormalCube(IBlockAccess aWorld, int aX, int aY, int aZ)  {return mNormalCube;}
	@Override public final String getUnlocalizedName() {return mNameInternal;}
	@Override public final String getLocalizedName() {return StatCollector.translateToLocal(mNameInternal + ".name");}
	@Override public final String getHarvestTool(int aMaterialToolQuality) {return mTool;}
	@Override public final int getHarvestLevel(int aMaterialToolQuality) {return (int)UT.Code.bind(mHarvestLevelMinimum, mHarvestLevelMaximum, mHarvestLevelOffset + aMaterialToolQuality);}
	@Override public final boolean canBeReplacedByLeaves(IBlockAccess aWorld, int aX, int aY, int aZ) {return false;}
	@Override public final boolean hasTileEntity(int aMeta) {return true;}
	@Override public final boolean canSilkHarvest() {return false;}
	@Override public final TileEntity createNewTileEntity(World aWorld, int aMeta) {return null;}
	@Override public final TileEntity createTileEntity(World aWorld, int aMeta) {return null;}
    @Override public final void getSubBlocks(Item aItem, CreativeTabs aCreativeTab, List aList) {/**/}
	@Override public final ITexture getTexture(int aRenderPass, byte aSide, short aMetaData) {return null;}
	@Override public final ITexture getTexture(int aRenderPass, byte aSide, IBlockAccess aWorld, int aX, int aY, int aZ) {return null;}
	@Override public final boolean setBlockBounds(int aRenderPass, byte aSide, short aMetaData) {return false;}
	@Override public final boolean setBlockBounds(int aRenderPass, byte aSide, IBlockAccess aWorld, int aX, int aY, int aZ) {return false;}
	@Override public final int getRenderPasses(byte aSide, short aMetaData) {return 0;}
	@Override public final int getRenderPasses(byte aSide, IBlockAccess aWorld, int aX, int aY, int aZ) {return 0;}
	@Override public final IRenderedBlockObject passRenderingToObject(byte aSide, short aMetaData) {return null;}
	@Override public final void registerBlockIcons(IIconRegister aIconRegister) {/**/}
    @Override public final IIcon getIcon(IBlockAccess aWorld, int aX, int aY, int aZ, int aSide) {return null;}
    @Override public final IIcon getIcon(int aSide, int aMetaData) {return null;}
	@Override public final int getRenderType() {return RendererBlockTextured.INSTANCE.mRenderID;}
	@Override public final IRenderedBlockObject passRenderingToObject(IBlockAccess aWorld, int aX, int aY, int aZ) {TileEntity tTileEntity = aWorld.getTileEntity(aX, aY, aZ); return tTileEntity instanceof IRenderedBlockObject ? (IRenderedBlockObject)tTileEntity : null;}
	@Override public final void onNeighborChange(IBlockAccess aWorld, int aX, int aY, int aZ, int aTileX, int aTileY, int aTileZ) {if (!LOCK) {LOCK = true; TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); if (aTileEntity instanceof ITileEntity) ((ITileEntity)aTileEntity).onAdjacentBlockChange(aTileX, aTileY, aTileZ); LOCK = false;}}
    @Override public final void onNeighborBlockChange(World aWorld, int aX, int aY, int aZ, Block aBlock) {if (!LOCK) {LOCK = true; TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); if (aTileEntity instanceof ITileEntity) ((ITileEntity)aTileEntity).onAdjacentBlockChange(aX, aY, aZ); LOCK = false;} aWorld.scheduleBlockUpdate(aX, aY, aZ, this, tickRate(aWorld));}
    @Override public final boolean onBlockEventReceived(World aWorld, int aX, int aY, int aZ, int aID, int aData) {TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ); return aTileEntity == null || aTileEntity.receiveClientEvent(aID, aData);}
}