package gregapi.item.prefixitem;

import static gregapi.data.CS.SHOW_HIDDEN_MATERIALS;
import static gregapi.data.CS.SHOW_HIDDEN_PREFIXES;
import static gregapi.data.CS.W;
import gregapi.GT_API;
import gregapi.code.ItemStackContainer;
import gregapi.data.MT;
import gregapi.data.TD;
import gregapi.item.CreativeTab;
import gregapi.item.IItemUpdatable;
import gregapi.lang.LanguageHandler;
import gregapi.oredict.OreDictManager;
import gregapi.oredict.OreDictMaterial;
import gregapi.oredict.OreDictPrefix;
import gregapi.render.IIconContainer;
import gregapi.util.UT;

import java.util.HashSet;
import java.util.List;

import net.minecraft.client.renderer.texture.IIconRegister;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.util.IIcon;
import net.minecraft.world.World;
import cpw.mods.fml.common.registry.GameRegistry;

/**
 * @author Gregorius Techneticies
 */
public class PrefixItem extends Item implements Runnable, IItemUpdatable {
	public final String mNameInternal;
	public final OreDictPrefix mPrefix;
	
	/**
	 * 
	 * @param aModIDOwner the ID of the owning Mod. DO NOT INSERT ANY GREGTECH MODID!!!
	 * @param aModIDTextures the ID of the Texture providing Mod (for the "ModID:TextureName" thing)
	 * @param aNameInternal the internal Name of this Item. DO NOT START YOUR UNLOCALISED NAME WITH "gt."!!!
	 * @param aPrefix the OreDictPrefix corresponding to this Item.
	 */
	public PrefixItem(String aModIDOwner, String aModIDTextures, String aNameInternal, OreDictPrefix aPrefix) {
		super();
		mPrefix = aPrefix;
		mNameInternal = aNameInternal;
		GameRegistry.registerItem(this, mNameInternal, aModIDOwner);
		
		setMaxDamage(0);
        setHasSubtypes(true);
		mPrefix.addTextureSet(aModIDTextures, true);
		LanguageHandler.add(mNameInternal+"."+0+".name", getLocalName(mPrefix, MT.Empty));
		LanguageHandler.add(mNameInternal+"."+W+".name", "Any Sub-Item of this one"); // Local Name for the WildcardItem Variant.
        mPrefix.mRegisteredItems.add(new ItemStackContainer(this, 1, W)); // this optimises some processes by decreasing the size of the Set.
        
        if (mPrefix.mCreativeTab == null) mPrefix.mCreativeTab = new CreativeTab(mPrefix.mNameInternal, mPrefix.mNameCategory, this, W);
		setCreativeTab(mPrefix.mCreativeTab);
		
		GT_API.sAPIInitBefore.add(this);
	}
	
	/** This ensures, that all Materials are registered at the time this Item registers to the OreDictionary. */
	@Override
	public void run() {
		boolean tUnificationAllowed = (mPrefix.contains(TD.Prefix.UNIFICATABLE) && !mPrefix.contains(TD.Prefix.UNIFICATABLE_RECIPES));
		for (short i = 0; i < OreDictMaterial.MATERIAL_ARRAY.length; i++) if (OreDictMaterial.MATERIAL_ARRAY[i] != null && mPrefix.isGeneratingItem(OreDictMaterial.MATERIAL_ARRAY[i])) {
			ItemStack tStack = new ItemStack(this, 1, i);
			UT.Stacks.update(tStack);
			LanguageHandler.add(getUnlocalizedName(tStack) + ".name", getLocalName(mPrefix, OreDictMaterial.MATERIAL_ARRAY[i]));
			if (tUnificationAllowed) OreDictManager.INSTANCE.addTarget(mPrefix, OreDictMaterial.MATERIAL_ARRAY[i], tStack); else OreDictManager.INSTANCE.registerOre(mPrefix, OreDictMaterial.MATERIAL_ARRAY[i], tStack);
		}
	}
	
	@Override
    public void getSubItems(Item var1, CreativeTabs aCreativeTab, List aList) {
        if (SHOW_HIDDEN_PREFIXES || !mPrefix.contains(TD.Creative.HIDDEN)) for (int i = 0; i < OreDictMaterial.MATERIAL_ARRAY.length; i++) if (mPrefix.isGeneratingItem(OreDictMaterial.MATERIAL_ARRAY[i])) if (SHOW_HIDDEN_MATERIALS || !OreDictMaterial.MATERIAL_ARRAY[i].contains(TD.Creative.HIDDEN)) {
			ItemStack tStack = new ItemStack(this, 1, i);
			updateItemStack(tStack);
        	aList.add(tStack);
        }
        if (aList.isEmpty()) try {codechicken.nei.api.API.hideItem(new ItemStack(this, 1, W));} catch(Throwable e) {/**/}
    }
	
	@Override
    public int getRenderPasses(int metadata) {
        return 2;
    }
    
	@Override
    public boolean requiresMultipleRenderPasses() {
        return mPrefix.mIconIndexItem >= 0;
    }
	
	@Override
    public void registerIcons(IIconRegister aIconRegister) {
		if (mPrefix.mIconIndexItem >= 0) {
			MT.NULL.mTextureSetsItems.get(mPrefix.mIconIndexItem).registerIcons(aIconRegister);
			HashSet<IIconContainer> tSet = new HashSet<IIconContainer>();
			for (int i = 0; i < OreDictMaterial.MATERIAL_ARRAY.length; i++) if (OreDictMaterial.MATERIAL_ARRAY[i] != null && OreDictMaterial.MATERIAL_ARRAY[i].mTextureSetsItems != null) {
				IIconContainer tIcon = OreDictMaterial.MATERIAL_ARRAY[i].mTextureSetsItems.get(mPrefix.mIconIndexItem);
				if (!tSet.contains(tIcon)) {tIcon.registerIcons(aIconRegister); tSet.add(tIcon);}
			}
		}
    }
	
	@Override
    public int getSpriteNumber() {
        return 1;
    }
	
	@Override
    public IIcon getIconIndex(ItemStack aStack) {
        return getIconFromDamageForRenderPass(aStack.getItemDamage(), 0);
    }
	
	@Override
    public IIcon getIconFromDamage(int aMetaData) {
    	return getIconFromDamageForRenderPass(aMetaData, 0);
    }
	
	@Override
    public IIcon getIcon(ItemStack aStack, int aRenderPass) {
        return getIconFromDamageForRenderPass(aStack.getItemDamage(), aRenderPass);
    }
	
	@Override
    public IIcon getIcon(ItemStack aStack, int aRenderPass, EntityPlayer aPlayer, ItemStack aUsedStack, int aUseRemaining) {
        return getIconFromDamageForRenderPass(aStack.getItemDamage(), aRenderPass);
    }
	
    @Override
    public IIcon getIconFromDamageForRenderPass(int aMetaData, int aRenderPass) {
		if (mPrefix.mIconIndexItem >= 0) {
			if (UT.Code.exists(aMetaData, OreDictMaterial.MATERIAL_ARRAY) && OreDictMaterial.MATERIAL_ARRAY[aMetaData].mTextureSetsItems != null)
			return OreDictMaterial.MATERIAL_ARRAY[aMetaData]	.mTextureSetsItems.get(mPrefix.mIconIndexItem).getIcon(aRenderPass);
			return MT.NULL										.mTextureSetsItems.get(mPrefix.mIconIndexItem).getIcon(aRenderPass);
		}
        return null;
    }
	
	@Override
    public int getColorFromItemStack(ItemStack aStack, int aRenderPass) {
		if (aRenderPass == 0) {
			int aMetaData = aStack.getItemDamage();
			if (UT.Code.exists(aMetaData, OreDictMaterial.MATERIAL_ARRAY)) return UT.Code.getRGBInt(OreDictMaterial.MATERIAL_ARRAY[aMetaData].mRGBa[mPrefix.mState]);
		}
		return 16777215;
    }
	
	@Override public final Item setUnlocalizedName(String aName) {return this;}
	@Override public final String getUnlocalizedName() {return mNameInternal;}
	@Override public final String getUnlocalizedName(ItemStack aStack) {return getHasSubtypes()?mNameInternal+"."+getDamage(aStack):mNameInternal;}
	@Override public final boolean hasContainerItem(ItemStack aStack) {return getContainerItem(aStack) != null;}
	@Override public ItemStack getContainerItem(ItemStack aStack) {return null;}
	@Override public void updateItemStack(ItemStack aStack) {int aDamage = aStack.getItemDamage(); if (UT.Code.exists(aDamage, OreDictMaterial.MATERIAL_ARRAY)) {OreDictMaterial aMaterial = OreDictMaterial.MATERIAL_ARRAY[aDamage]; if (aMaterial != aMaterial.mTargetRegistration) aStack.setItemDamage(aMaterial.mTargetRegistration.mID);}}
	@Override public void onCreated(ItemStack aStack, World aWorld, EntityPlayer aPlayer) {updateItemStack(aStack);}
	@Override public void addInformation(ItemStack aStack, EntityPlayer aPlayer, List aList, boolean aF3_H) {/**/}
	@Override public boolean isBookEnchantable(ItemStack aStack, ItemStack aBook) {return false;}
	@Override public boolean getIsRepairable(ItemStack aStack, ItemStack aMaterial) {return false;}
	@Override public int getItemEnchantability() {return 0;}
	@Override public int getItemStackLimit(ItemStack aStack) {return mPrefix.mDefaultStackSize;}
	
	/** @return the Local Name for this Item depending on Prefix and Material. */
	public String getLocalName(OreDictPrefix aPrefix, OreDictMaterial aMaterial) {
		return LanguageHandler.getLocalName(aPrefix, aMaterial);
	}
}